#include <stdarg.h>

#include "doomtype.h"
#include "doomstat.h"
#include "d_player.h"
#include "s_sound.h"
#include "gi.h"
#include "d_net.h"
#include "g_game.h"
#include "p_local.h"
#include "sv_main.h"
#include "i_system.h"
#include "c_console.h"
#include "c_dispatch.h"
#include "m_argv.h"
#include "m_random.h"
#include "s_sndseq.h"
#include "v_video.h"

#include "w_wad.h"

int     lastreliablenum;
short   parse_cl;

client_t clients[MAXPLAYERS];

int       maxclients;
int       max_players;
int		  server_outdated=-10;
int		  controlledbytes=0; // these are bytes sent to the server that
							 // the player has control over.
client_t  tempclient;

char	server_wads[4][32];	// ZDaemon will support up to 4 wad files
						// to report to the launcher

#define CLIENT_TIMEOUT 65 // 65 seconds

// Different welcome strings
char welcomestring[] = "Version 0.99 Server\n";
char welcomestring_teamplay[] = "Version 0.99 Teamplay Server\n";
int  gametime;

void SV_UpdateConsolePlayer(client_t* cl);
void SV_Fraglogfile (void);
void SV_Logconnection (char *text);
int SV_ChooseTeamForPlayer(void);
void P_SpawnPlayer (mapthing2_t* mthing);

CVAR (rcon_password, "", CVAR_SERVERINFO|CVAR_ARCHIVE)
CVAR (cfg_activated, "-10", CVAR_SERVERINFO)
CVAR (max_rate,      "8192", CVAR_SERVERINFO|CVAR_ARCHIVE)
CVAR (mapcheckdifference, "5", CVAR_SERVERINFO|CVAR_ARCHIVE)
CVAR (floodrate, "1000", CVAR_SERVERINFO|CVAR_ARCHIVE)
CVAR (motd, "Welcome to ZDaemon<br>0.95 server<br><br><br>Happy fragging!", CVAR_SERVERINFO|CVAR_ARCHIVE)

BEGIN_COMMAND (fraglog)
{
	SV_Fraglogfile();
}
END_COMMAND (fraglog)

BEGIN_COMMAND (wads)
{
	Printf(PRINT_HIGH, "\n");
	for (int i=0; i<4; i++)
	{
		if (!server_wads[i])
			continue;

		Printf(PRINT_HIGH, "%s\n", server_wads[i]);
	}

	Printf(PRINT_HIGH, "\n");
}
END_COMMAND(wads)

//BEGIN_COMMAND(crash)
//{
//	if (argc <= 0)
//		return;
//	
//	int t = 1/atoi(argv[1]);
//}
//END_COMMAND(crash)


//
// [NightFang] - Add this...some *unix compiliers might not support it
//
void stringupr(char *s)
{
	while (*s)
	*s++ = toupper (*s);
}


//
// SV_CheckConfig - Checks for a server config file 'zserver.cfg'
//
void SV_CheckConfig(void)
{
	if (!FileExists("zserver.cfg"))
	{
		I_FatalError("Unable to load- zserver.cfg\n");
		return;
	}

	AddCommandString("exec zserver.cfg");

	if (cfg_activated.value <= 0)
	{
		I_FatalError("\nYou need to edit your zserver.cfg file and activate it\n");
		return;
	}
}




//
// SV_InitNetwork
//
void SV_InitNetwork (void)
{
    char      *v;
	int		  num_bans;

	netgame = false;  // for old network code
	multiplayer = true;

// Clear server wads
//	memset(server_wads, 0, sizeof(server_wads));

// Must have CFG set in order to use this
	SV_CheckConfig();
 
	v = Args.CheckValue ("-port");
    if (v)
    {
       localport = atoi (v);
       Printf (PRINT_HIGH, "using alternate port %i\n", localport);
    }
	else 
	   localport = SERVERPORT;
	
	// set up a socket and net_message buffer
	InitNetCommon();

	// determine my name & address
	// NET_GetLocalAddress ();

	Printf(PRINT_HIGH, "UDP Initialized\n");

	// Max amount of connectiosn that can be made
	v = Args.CheckValue ("-maxclients");
    if (v)
    {
		maxclients = atoi(v);
		if (maxclients < 0)
			maxclients = 0;
		if (maxclients > MAXPLAYERS)
			maxclients = MAXPLAYERS;
	}
	else
		maxclients = MAXPLAYERS;

	
	// Max amount of players that can actually join
	v = Args.CheckValue ("-maxplayers");
    if (v)
    {
		max_players = atoi(v);
		if (max_players < 0)
			max_players = 0;
		if (max_players > maxclients)
			maxclients = maxclients;
	}
	else
		max_players = maxclients;

	// Let the admin know what slots are open
	Printf(PRINT_HIGH, "%d connections total. %d can be used for players.\n", 
		maxclients, max_players);

	

	// initialize clients
	for (int i=0; i<maxclients; i++)
    {
		clients[i].netbuf.data = new byte[MAX_UDP_PACKET];
		clients[i].netbuf.maxsize = MAX_UDP_PACKET;
		SZ_Clear(&clients[i].netbuf);
	   
		clients[i].reliablebuf.data = new byte[MAX_UDP_PACKET];
		clients[i].reliablebuf.maxsize = MAX_UDP_PACKET;
		SZ_Clear(&clients[i].reliablebuf);

		clients[i].relpackets.data = new byte[1024*50];
		clients[i].relpackets.maxsize = 1024*50; // 50 KB

		clients[i].cls = cs_free;
	}

	tempclient.reliablebuf.data = new byte[MAX_UDP_PACKET];
    tempclient.reliablebuf.maxsize = MAX_UDP_PACKET;

	gametime = I_GetTime ();

// [NightFang] - Load the previously recorded bans into file
	num_bans = SV_ReadBans(BANFILE);
	Printf(PRINT_HIGH, "Loaded: %d server bans.\n", num_bans);
	
	SV_InitMaster();	
}


int SV_GetFreeClient(void)
{
	int i;

	for (i=0; i < maxclients; i++)
	{
		if (clients[i].cls == cs_free)
			return i;
	}
	
	return -1;
}

short SV_FindClientByAddr(void)
{
    int i;

	for (i=0; i<MAXPLAYERS; i++)
	{
	   if (clients[i].cls == cs_free)
	       continue;
		
	   if (NET_CompareAdr(clients[i].address, net_from) )
	       return i;
	}
	
	return -1;
}

//
// SV_CheckTimeouts
// If a packet has not been received from a client in CLIENT_TIMEOUT
// seconds, drop the conneciton.
//
void SV_CheckTimeouts (void)
{
	for (int i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		if (gametic - cl->last_received == CLIENT_TIMEOUT*35)
		{
			parse_cl = i;
		    SV_DropClient();
		}
	}
}

//
// SV_GetPackets
//
void SV_GetPackets (void)
{
	int	packet_size;

    while (packet_size = NET_GetPacket() )
    {
        // packet loss emulation :)
        //if (P_Random(pr_misc) > 150)
		//	continue;

        parse_cl = SV_FindClientByAddr();

        if (parse_cl == -1) // no client with net_from address
		{
    	    // apparently, someone is trying to connect
    	    if (gamestate == GS_LEVEL)
				SV_GetConnection();
            continue;
        }

		clients[parse_cl].last_received = gametic;
		//clients[parse_cl].bytesin += packet_size;

		if (clients[parse_cl].cls != cs_spawned)
		{
			int		msg = MSG_ReadLong();

			if (msg == S2C_CHALLENGE)
				SV_GetChallenge(10); // use parse_cl
			if (msg == S2C_VERSION)
				SV_ConnectClient();
			if (msg == S2C_CONNECTION)
				SV_SpawnClient();
			continue;
		}

		/*if (clients[parse_cl].cls == cs_challenge)
		{
			SV_ConnectClient();
			continue;
		}

		if (clients[parse_cl].cls == cs_connected)
		{
			SV_SpawnClient();
			continue;
		}*/

		SV_ParseCommands();
    }
}

//
// SV_Sound
//
void SV_Sound (AActor *mo, byte channel, char *name, byte attenuation)
{
	int        sfx_id;
	client_t  *cl;
	int        i;

	if (*name == '*')
	{
		for (i=0; i<MAXPLAYERS; i++)
		{
			cl = &clients[i];

			if (cl->cls < cs_spawned)
				continue;
		
			SV_CheckBuffer(cl, &cl->netbuf, 6+sizeof(name));
			MSG_WriteByte (&cl->netbuf, svc_startskinnedsound);
			if (mo == NULL)
				MSG_WriteShort (&cl->netbuf, 0);
			else
				MSG_WriteShort (&cl->netbuf, mo->netid);
			MSG_WriteByte (&cl->netbuf, channel);
			MSG_WriteString (&cl->netbuf, name);
			MSG_WriteByte (&cl->netbuf, attenuation);
		}

		return;

		/*// Sexed sound
		char nametemp[128];
		const char templat[] = "player/%s/%s";
		const char *genders[] = { "male", "female", "cyborg" };
		player_t *player;

		sfx_id = -1;
		if ( (player = mo->player) ) 
		{
			sprintf (nametemp, templat, skins[player->userinfo.skin].name, name + 1);
			sfx_id = S_FindSound (nametemp);
			if (sfx_id == -1) 
			{
				sprintf (nametemp, templat, genders[player->userinfo.gender], name + 1);
				sfx_id = S_FindSound (nametemp);
			}
		}
		if (sfx_id == -1) 
		{
			sprintf (nametemp, templat, "male", name + 1);
			sfx_id = S_FindSound (nametemp);
		}*/
	} 
	else
		sfx_id = S_FindSound (name);


	//if (sfx_id > 255 || sfx_id < 0)
	if (!sfx_id)
	{
		Printf (PRINT_HIGH, "SV_Sound: range error. Sfx_id = %d\n", sfx_id);
		return;
	}

	for (i=0; i<MAXPLAYERS; i++)
	{
		cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		
		SV_CheckBuffer(cl, &cl->netbuf, 6);
		MSG_WriteByte (&cl->netbuf, svc_startsound);
		MSG_WriteShort (&cl->netbuf, mo->netid);
		MSG_WriteByte (&cl->netbuf, channel);
		MSG_WriteByte (&cl->netbuf, sfx_id);
		MSG_WriteByte (&cl->netbuf, attenuation);
	}

}


void SV_SexedSound (AActor *mo, byte channel, char *name, byte attenuation)
{
//	int sfx_id;

	if (*name == '*')
	{
		for (int i=0; i<MAXPLAYERS; i++)
		{
			client_t* cl = &clients[i];

			if (cl->cls < cs_spawned)
				continue;
		
			SV_CheckBuffer(cl, &cl->netbuf, 6+sizeof(name));
			MSG_WriteByte (&cl->netbuf, svc_startskinnedsound);
			if (mo == NULL)
				MSG_WriteShort (&cl->netbuf, 0);
			else
				MSG_WriteShort (&cl->netbuf, mo->netid);
			MSG_WriteByte (&cl->netbuf, channel);
			MSG_WriteString (&cl->netbuf, name);
			MSG_WriteByte (&cl->netbuf, attenuation);
		}

		return;
	
	} 
}

//
// SV_SoundAvoidCl
// Sends a sound to clients, but doesn't send it to client 'player'.
//
void SV_SoundAvoidCl (player_t *player, AActor *mo, byte channel, char *name, byte attenuation)
{
	int        i;
	int        sfx_id;
	client_t  *cl;

	if (*name == '*')
	{
		for (i=0; i<MAXPLAYERS; i++)
		{
			cl = &clients[i];

			if (cl->cls < cs_spawned)
				continue;
		
			SV_CheckBuffer(cl, &cl->netbuf, 6+sizeof(name));
			MSG_WriteByte (&cl->netbuf, svc_startskinnedsound);
			if (mo == NULL)
				MSG_WriteShort (&cl->netbuf, 0);
			else
				MSG_WriteShort (&cl->netbuf, mo->netid);
			MSG_WriteByte (&cl->netbuf, channel);
			MSG_WriteString (&cl->netbuf, name);
			MSG_WriteByte (&cl->netbuf, attenuation);
		}

		return;

		// Sexed sound
		/*char nametemp[128];
		const char templat[] = "player/%s/%s";
		const char *genders[] = { "male", "female", "cyborg" };
		player_t *player;

		sfx_id = -1;
		if ( (player = mo->player) ) 
		{
			sprintf (nametemp, templat, skins[player->userinfo.skin].name, name + 1);
			sfx_id = S_FindSound (nametemp);
			if (sfx_id == -1) 
			{
				sprintf (nametemp, templat, genders[player->userinfo.gender], name + 1);
				sfx_id = S_FindSound (nametemp);
			}
		}
		if (sfx_id == -1) 
		{
			sprintf (nametemp, templat, "male", name + 1);
			sfx_id = S_FindSound (nametemp);
		}*/
	} 
	else
		sfx_id = S_FindSound (name);

	//if (sfx_id > 255 || sfx_id < 0)
	if (!sfx_id)
	{
		Printf (PRINT_HIGH, "SV_Sound: range error. Sfx_id = %d\n", sfx_id);
		return;
	}

    for (i=0; i<MAXPLAYERS; i++)
    {
		cl = &clients[i];

		if (cl->cls < cs_spawned || player == &players[i])
			continue;

		SV_CheckBuffer(cl, &cl->netbuf, 6);
		MSG_WriteByte (&cl->netbuf, svc_startsound);
		MSG_WriteShort (&cl->netbuf, mo->netid);
		MSG_WriteByte (&cl->netbuf, channel);
		MSG_WriteByte (&cl->netbuf, sfx_id);
		MSG_WriteByte (&cl->netbuf, attenuation);
    }
}


void SV_Sound (fixed_t x, fixed_t y, byte channel, char *name, byte attenuation)
{
	int        sfx_id;
	client_t  *cl;
	int        i;

	sfx_id = S_FindSound (name);

	//if (sfx_id > 255 || sfx_id < 0)
	if (!sfx_id)
	{
		Printf (PRINT_HIGH, "SV_Sound: range error. Sfx_id = %d\n", sfx_id);
		return;
	}

	for (i=0; i<MAXPLAYERS; i++)
	{
		cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		SV_CheckBuffer(cl, &cl->netbuf, 12);
		MSG_WriteByte (&cl->netbuf, svc_soundorigin);
		MSG_WriteLong (&cl->netbuf, x);
		MSG_WriteLong (&cl->netbuf, y);
		MSG_WriteByte (&cl->netbuf, channel);
		MSG_WriteByte (&cl->netbuf, sfx_id);
		MSG_WriteByte (&cl->netbuf, attenuation);
	}

}


//
// SV_StartSequence
//
void SV_StartSequence (sector_t *sector, int sequence, seqtype_t type)
{
	int sec, i;

	for (sec=0; sec<numsectors; sec++)
		if (sector == &sectors[sec])
			break;

	for (i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		SV_CheckBuffer(cl, &cl->reliablebuf, 6);
		MSG_WriteByte (&cl->reliablebuf, svc_startsectorsequence);
		MSG_WriteShort (&cl->reliablebuf, sec);
		MSG_WriteShort (&cl->reliablebuf, sequence);
		MSG_WriteByte (&cl->reliablebuf, (byte)type);
	}
}

//
// SV_StopSequence
//
void SV_StopSequence(sector_t *sector)
{
	int sec, i;

	for (sec=0; sec<numsectors; sec++)
		if (sector == &sectors[sec])
			break;

	for (i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		SV_CheckBuffer(cl, &cl->reliablebuf, 3);
		MSG_WriteByte (&cl->reliablebuf, svc_stopsectorsequence);
		MSG_WriteShort (&cl->reliablebuf, sec);
	}
}


//
// SV_UpdateFrags
//
void SV_UpdateFrags (player_t *player)
{
    int  playernum;

    for (playernum=0; playernum<MAXPLAYERS; playernum++)
    {
        if (player == &players[playernum])
            break;
    }

    for (int i=0; i<MAXPLAYERS; i++)
    {
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

        SV_CheckBuffer(cl, &cl->reliablebuf, 4);
		MSG_WriteByte (&cl->reliablebuf, svc_updatefrags);
		MSG_WriteByte (&cl->reliablebuf, playernum);
        MSG_WriteShort (&cl->reliablebuf, player->fragcount);
    }
}

//
// SV_SendUserInfo
//
void SV_SendUserInfo (int player, client_t* cl)
{
	player_t *p = &players[player];

	SV_CheckBuffer(cl, &cl->reliablebuf, 11+sizeof(p->userinfo.netname)+sizeof(p->userinfo.team)+sizeof(skins[p->userinfo.skin].name));
	MSG_WriteByte (&cl->reliablebuf, svc_userinfo);
	MSG_WriteByte (&cl->reliablebuf, player);
    MSG_WriteString (&cl->reliablebuf, p->userinfo.netname);
    MSG_WriteString (&cl->reliablebuf, p->userinfo.team);
    MSG_WriteLong (&cl->reliablebuf, p->userinfo.gender);
    MSG_WriteLong (&cl->reliablebuf, p->userinfo.color);
	MSG_WriteString(&cl->reliablebuf, skins[p->userinfo.skin].name);

	if (p->userinfo.neverswitch)
		MSG_WriteByte (&cl->reliablebuf, 10);
	else
		MSG_WriteByte (&cl->reliablebuf, 0);

}


void SV_SetupUserInfo(void)
{
    player_t	*p;
	char		*str;
	char		*skin;
	
    p = &players[parse_cl];

	// 
	str = MSG_ReadString();
	if (strlen(str) > MAXPLAYERNAME)
		str[MAXPLAYERNAME] = 0;

	// let everyone else know
	stringupr(str);

	// the two names are different
	if (strcmp(p->userinfo.netname, str))
	{
		if (clients[parse_cl].cls == cs_spawned && strlen(p->userinfo.netname) > 0)
			SV_BroadcastPrintf(PRINT_HIGH, "%s is now known as %s.\n", p->userinfo.netname, str);
	}

	strcpy(p->userinfo.netname, str);
	
	str = MSG_ReadString();
	if (strlen(str) > MAXPLAYERNAME)
		str[MAXPLAYERNAME] = 0;
	
	
    strcpy(p->userinfo.team, str);

    p->userinfo.gender = MSG_ReadLong();
    p->userinfo.color = MSG_ReadLong();
    p->userinfo.aimdist = MSG_ReadLong();

	skin = MSG_ReadString();

	// Make sure the skin is valid
	p->userinfo.skin = R_FindSkin(skin);

	if (MSG_ReadByte() > 0)
		p->userinfo.neverswitch = true;
	else
		p->userinfo.neverswitch = false;


	// send to other players
	for (int i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)// || i==parse_cl)
			continue;

		SV_SendUserInfo(parse_cl, cl);
	}

	clients[parse_cl].bytesin += sizeof(str)+12+sizeof(skin)+1;
}

void SV_UpdateFloorpic(int s, client_t* cl)
{
	char	str[32];

	// Update sector floors
	if (sectors[s].tag)
	{
		W_GetLumpName (str, (firstflat + sectors[s].floorpic));

		SV_CheckBuffer(cl, &cl->reliablebuf, sizeof(str)+3);

		SV_CheckBuffer(cl, &cl->reliablebuf, 5);
		MSG_WriteByte(&cl->reliablebuf, svc_floorpic);
		MSG_WriteShort(&cl->reliablebuf, s);
		MSG_WriteString(&cl->reliablebuf, str);
	}
}




//
// SV_ClientFullUpdate
//
void SV_ClientFullUpdate (client_t *cl)
{
	AActor *mo;
	int     i;
	int		playerstate;

	// send players info to the client
	for (i=0; i<MAXPLAYERS; i++)
	{
		if (clients[i].cls < cs_spawned || cl == &clients[i])
			continue;

		if (players[i].playerstate == PST_WATCHING || players[i].spectator > 0)
		{ playerstate = PST_WATCHING; }
		else
		{ playerstate = PST_REBORN; }

		SV_CheckBuffer(cl, &cl->reliablebuf, 25);
		MSG_WriteByte(&cl->reliablebuf, svc_spawnplayer);
		MSG_WriteByte(&cl->reliablebuf, i);
		MSG_WriteByte (&cl->reliablebuf, playerstate);
		MSG_WriteByte (&cl->reliablebuf, 10);
		MSG_WriteByte (&cl->reliablebuf, players[i].spectator);
		MSG_WriteShort(&cl->reliablebuf, players[i].mo->netid);
		MSG_WriteShort(&cl->reliablebuf, players[i].mo->handle_id);
		MSG_WriteLong(&cl->reliablebuf, players[i].mo->angle);
		MSG_WriteLong(&cl->reliablebuf, players[i].mo->x);
		MSG_WriteLong(&cl->reliablebuf, players[i].mo->y);
		MSG_WriteLong(&cl->reliablebuf, players[i].mo->z);

		SV_SendUserInfo(i, cl);
	}

	// update frags
	for (i=0; i<MAXPLAYERS; i++)
	{
		if (!playeringame[i])
			continue;

		SV_CheckBuffer(cl, &cl->reliablebuf, 4);
		MSG_WriteByte(&cl->reliablebuf, svc_updatefrags);
		MSG_WriteByte(&cl->reliablebuf, i);
		MSG_WriteShort(&cl->reliablebuf, players[i].fragcount);
	}

	// update things
	TThinkerIterator<AActor> iterator;
    while ( (mo = iterator.Next() ) )
    {
		if (!mo->netid)
			continue;

        if (mo->type != MT_PLAYER)
		{
			SV_CheckBuffer(cl, &cl->reliablebuf, 21);
			MSG_WriteByte(&cl->reliablebuf, svc_spawnmobj);
			MSG_WriteLong(&cl->reliablebuf, mo->x);
			MSG_WriteLong(&cl->reliablebuf, mo->y);
			MSG_WriteLong(&cl->reliablebuf, mo->z);
			MSG_WriteShort(&cl->reliablebuf, mo->type);
			MSG_WriteShort(&cl->reliablebuf, mo->netid);
			MSG_WriteShort(&cl->reliablebuf, mo->handle_id);

			// a monster
			if (mo->type == MT_SKULL || (mo->flags & MF_COUNTKILL) )
			{
				SV_CheckBuffer(cl, &cl->reliablebuf, 7);
				MSG_WriteByte(&cl->reliablebuf, svc_mobjangle);
				MSG_WriteShort(&cl->reliablebuf, mo->netid);
				MSG_WriteLong(&cl->reliablebuf, mo->angle);
			}
		}

		// corpses
		if ( mo->health <= 0 && (mo->type == MT_PLAYER 
			|| (mo->flags & MF_COUNTKILL) ) )
		{
			SV_CheckBuffer(cl, &cl->reliablebuf, 4);
			MSG_WriteByte(&cl->reliablebuf, svc_corpse);
			MSG_WriteShort(&cl->reliablebuf, mo->netid);
			MSG_WriteByte (&cl->reliablebuf, mo->frame);
		}

		//if (cl->reliablebuf.cursize >= 4096)
		//	SV_SendPacket(cl);
    }

	// update sectors
	for (int s=0; s<numsectors; s++)
	{ 
		sector_t* sec = &sectors[s];

	    //if (sec->moveable)
		//{
		SV_CheckBuffer(cl, &cl->reliablebuf, 9);
		MSG_WriteByte (&cl->reliablebuf, svc_sector);
		MSG_WriteShort (&cl->reliablebuf, s);
		MSG_WriteShort (&cl->reliablebuf, sec->floorheight>>FRACBITS);
		MSG_WriteShort (&cl->reliablebuf, sec->ceilingheight>>FRACBITS);
		MSG_WriteShort (&cl->reliablebuf, sec->lightlevel);

					
		// Update sector floors pics
		SV_UpdateFloorpic(s, cl);

	}

	SV_SendPacket(cl);
}

//
// SV_LoadLevel
//
void SV_LoadLevel()
{
	player_t	*p;
	int	i;

	for (i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		p = &players[i];

		// send dmflags
		SV_CheckBuffer(cl, &cl->reliablebuf, 5);
		MSG_WriteByte(&cl->reliablebuf, svc_dmflags);
		MSG_WriteLong(&cl->reliablebuf, dmflags);

		// send skill
		SV_CheckBuffer(cl, &cl->reliablebuf, 2);
		MSG_WriteByte(&cl->reliablebuf, svc_gameskill);
		MSG_WriteByte(&cl->reliablebuf, (byte)gameskill.value);

		/*int team = SV_ChooseTeamForPlayer();

		if (team == 0 && teamplay.value)
		{
			p->userinfo.color = V_GetColorFromString (NULL, "ff 00 00");
			strcpy(p->userinfo.team, "RED");
		}
	
		if (team == 1 && teamplay.value)
		{
			p->userinfo.color = V_GetColorFromString (NULL, "00 00 ff");
			strcpy(p->userinfo.team, "BLUE");
		}

		// send to other players for teamplay
		if (teamplay.value)
		{
			for (int i=0; i<MAXPLAYERS; i++)
			{
				client_t *cl = &clients[i];
	
				if (cl->cls < cs_spawned)
					continue;
	
				SV_SendUserInfo(parse_cl, cl);
			}
		}*/

   
		// Send special settings like teamplay and deathmatch.
		MSG_WriteByte(&cl->reliablebuf, svc_specialsettings);
		MSG_WriteByte(&cl->reliablebuf, (byte)teamplay.value);
		MSG_WriteByte(&cl->reliablebuf, SV_ChooseTeamForPlayer());
		MSG_WriteByte(&cl->reliablebuf, (byte)deathmatch.value);

		// Force the client to load a new map.
		MSG_WriteByte(&cl->reliablebuf, svc_loadmap);
		MSG_WriteString(&cl->reliablebuf, level.mapname);		
	}
}


void SV_NewLevelClientUpdate()
{
	AActor *mo;
	int		i,j;

	for (i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		players[i].playerstate = PST_WATCHING;
		players[i].spectator = 10;
		players[i].cheats |= CF_NOTARGET;

		if (deathmatch.value)
			for (j=0; j<MAXPLAYERS; j++)
			{
				SV_CheckBuffer(cl, &cl->reliablebuf, 4);
				MSG_WriteByte(&cl->reliablebuf, svc_updatefrags);
				MSG_WriteByte(&cl->reliablebuf, j);
				MSG_WriteShort(&cl->reliablebuf, players[j].fragcount);
			}

		// update things
		TThinkerIterator<AActor> iterator;
		while ( (mo = iterator.Next() ) )
		{
			if (!mo->netid)
				continue;

			// a monster
			if (mo->type == MT_SKULL || (mo->flags & MF_COUNTKILL) )
			{
				SV_CheckBuffer(cl, &cl->reliablebuf, 7);
				MSG_WriteByte(&cl->reliablebuf, svc_mobjangle);
				MSG_WriteShort(&cl->reliablebuf, mo->netid);
				MSG_WriteLong(&cl->reliablebuf, mo->angle);
			}
		}

		SV_SendPacket(cl);
	}
}


//
// SV_ChooseTeamForPlayer
//
int SV_ChooseTeamForPlayer(void)
{
	int	reds=0;
	int	blues=0;
	player_t	*player;
	
	for(int i=0; i<MAXPLAYERS; i++)
	{
			if (clients[i].cls < cs_spawned)
			{ continue; }

			player = &players[i];

			if (!strcmp(player->userinfo.team, "RED")) // Red team
			{ reds++; }

			if (!strcmp(player->userinfo.team, "BLUE"))	// Blue team
			{ blues++;	}
	}

	if (reds > blues)
	{ return 1; }

	if (blues > reds)
	{ return 0; }

	if (blues == reds)
	{ return 0; }

	return 1;

}

void SV_GetConnection(void)
{
	char	str[128];
	int		msg = MSG_ReadLong();
	

	if (msg == S2C_CHALLENGE)
	{ SV_GetChallenge(0); }  // get new client
	else if (msg == LAUNCHER_CHALLENGE)
	{
		SV_SendServerInfo();  // for Launcher
		sprintf(str, "%s launcher request\n", NET_AdrToString(net_from));
		SV_Logconnection (str);
	}
	else if (msg == UPDATE_CHALLENGE)
	{
		server_outdated = MSG_ReadByte();
		if (server_outdated > 0)
		{ Printf(PRINT_HIGH, "Out dated server.\n"); }
		return;
	}

}

		


void SV_GetChallenge(int check)
{
	char	str[128];
	client_t *cl;
	int challenge = MSG_ReadLong();

	if (challenge == CHALLENGE)
	{
		sprintf(str, "%s game\n", NET_AdrToString(net_from));
		SV_Logconnection (str);
	}

	if (challenge != CHALLENGE)
	{	
		sprintf(str, "%s unknown request %d\n", NET_AdrToString(net_from),challenge);
		SV_Logconnection (str);
		return;
	}
		

	// We want to pull up a new client?
	if (check <= 0)
	{
		int n = SV_GetFreeClient();
		if (n==-1)  // a server is full
		{
			SZ_Clear(&tempclient.reliablebuf);
			MSG_WriteByte(&tempclient.reliablebuf, S2C_FULL);
			NET_SendPacket(tempclient.reliablebuf.cursize, tempclient.reliablebuf.data, net_from);
			return;
		}

		cl = &clients[n];
	}
	else
	{ cl = &clients[parse_cl]; }

	
	cl->address = net_from;
	cl->last_received = gametic;
	cl->cls = cs_challenge;

	SZ_Clear(&cl->reliablebuf);
	MSG_WriteByte(&cl->reliablebuf, S2C_CHALLENGE);
	NET_SendPacket(cl->reliablebuf.cursize, cl->reliablebuf.data, cl->address);
}


void SV_ConnectClient(void)
{
	client_t  *cl;

	cl = &clients[parse_cl];
	SZ_Clear(&cl->reliablebuf);

	short version = MSG_ReadShort();

	// Enforce any people who are banned
	if (SV_EnforceBans(cl) > 0) { return; }

	
	if (version != VERSION)
	{
		MSG_WriteByte(&cl->reliablebuf, svc_disconnect);
		NET_SendPacket(cl->reliablebuf.cursize, cl->reliablebuf.data, cl->address);
		cl->cls = cs_free;
		return;
	}

	MSG_WriteByte(&cl->reliablebuf, S2C_CONNECTION);
	NET_SendPacket(cl->reliablebuf.cursize, cl->reliablebuf.data, cl->address);
	cl->cls = cs_connected;

	Printf(PRINT_HIGH, "%s connection.\n", NET_AdrToString(net_from) );

	cl->reliable_bps = 0;
	cl->unreliable_bps = 0;
	cl->bytesin = 0;
	cl->lastcmdtic = 0;
	cl->allow_rcon = false;
	cl->lastsaytic = gametic;

	SZ_Clear(&cl->netbuf);
	SZ_Clear(&cl->reliablebuf);
	SZ_Clear(&cl->relpackets);

	memset(cl->packetseq, -1, sizeof(cl->packetseq) );
	memset(cl->packetbegin, 0, sizeof(cl->packetbegin) );
	memset(cl->packetsize, 0, sizeof(cl->packetsize) );
	cl->sequence = 0;
	cl->last_sequence = -1;
	cl->packetnum = 0;
}


// Function Prototype
void G_PlayerReborn (int player);

void SV_SpawnClient(void)
{
	client_t *cl = &clients[parse_cl];

	cl->cls = cs_spawned;

	// [NightFang] - Added count to keep better track of players 
	// and clients
	cl->count = parse_cl;

	// ZDaemon automatically set you in spectator mode when the 
	// player joins
	players[parse_cl].playerstate = PST_WATCHING;
	players[parse_cl].spectator = 10;
	players[parse_cl].needreborn = 10;
	players[parse_cl].cheats |= CF_NOTARGET;

	playeringame[parse_cl] = true;

	// send welcome message
	MSG_WriteByte (&cl->reliablebuf, svc_print);
	MSG_WriteByte (&cl->reliablebuf, PRINT_HIGH);
  
	// [NightFang] - Tell the player which type of server this is
	if (teamplay.value)
		MSG_WriteString (&cl->reliablebuf, welcomestring_teamplay);
	else
		MSG_WriteString (&cl->reliablebuf, welcomestring);

	MSG_WriteByte(&cl->reliablebuf, svc_sendmotd);
	MSG_WriteString(&cl->reliablebuf, (char *)motd.string);

	// send dmflags
	MSG_WriteByte(&cl->reliablebuf, svc_dmflags);
	MSG_WriteLong(&cl->reliablebuf, dmflags);

	// Send special settings like teamplay and deathmatch.
	MSG_WriteByte(&cl->reliablebuf, svc_specialsettings);
	MSG_WriteByte(&cl->reliablebuf, (byte)teamplay.value);
	MSG_WriteByte(&cl->reliablebuf, SV_ChooseTeamForPlayer());
	MSG_WriteByte(&cl->reliablebuf, (byte)deathmatch.value);


	// send consoleplayer number
	MSG_WriteByte(&cl->reliablebuf, svc_consoleplayer);
	MSG_WriteByte(&cl->reliablebuf, parse_cl);

	// send skill
	MSG_WriteByte(&cl->reliablebuf, svc_gameskill);
	MSG_WriteByte(&cl->reliablebuf, (byte)gameskill.value);



	
	memset (players[parse_cl].frags,0,sizeof(players[parse_cl].frags)); 
	players[parse_cl].fragcount = 0;


	// get client userinfo
	MSG_ReadByte();  // clc_userinfo
	SV_SetupUserInfo(); // send it to other players
						// if it's a team game, player's color
						// will be set to red or blue
	cl->rate = MSG_ReadLong(); // get rate value

	// Check the rate against the limit
	if (cl->rate > max_rate.value)
	{
		cl->rate = max_rate.value;
		Printf(PRINT_HIGH, "%s's rate re-adjusted to %d.\n", players[parse_cl].userinfo.netname, max_rate.value);
	}


	// send a map name
	MSG_WriteByte(&cl->reliablebuf, svc_loadmap);
	MSG_WriteString(&cl->reliablebuf, level.mapname);

	if (deathmatch.value)
		G_DeathMatchSpawnPlayer(parse_cl);
	else
		P_SpawnPlayer(&playerstarts[parse_cl]);

	SV_ClientFullUpdate(cl);

	SV_BroadcastPrintf (PRINT_HIGH, "%s has connected.\n", players[parse_cl].userinfo.netname);

	SV_UpdateFrags (&players[parse_cl]);
}


//
// SV_DisconnectClient
//
void SV_DisconnectClient(void)
{
	client_t *cl = &clients[parse_cl];

	playeringame[parse_cl] = false;
	cl->cls = cs_free;
    
	if (gametic - cl->last_received == CLIENT_TIMEOUT*35)
		SV_BroadcastPrintf (PRINT_HIGH, "%s timed out\n", players[parse_cl].userinfo.netname);
	else
        SV_BroadcastPrintf (PRINT_HIGH, "client %s disconnected\n", players[parse_cl].userinfo.netname);

	// tell others clients about it		
	for (int i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		MSG_WriteByte(&cl->reliablebuf, svc_disconnectclient);
		MSG_WriteByte(&cl->reliablebuf, parse_cl);
	}
	
	if (players[parse_cl].mo)
	{
		players[parse_cl].mo->Destroy();
		players[parse_cl].mo = NULL;
	}
	players[parse_cl].userinfo.netname[0] = '\0';
}


//
// SV_DropClient
// Called when the player is leaving the server unwillingly. 
//
void SV_DropClient(void)
{
	client_t *cl = &clients[parse_cl];

	MSG_WriteByte(&cl->reliablebuf, svc_disconnect);
	SV_SendPacket(cl);

	SV_DisconnectClient();

	//memset(players[parse_cl], 0, sizeof(players[parse_cl]));
}

//
// SV_SendReconnectSignal
// All clients will reconnect. Called when the server changes a map
//
void SV_SendReconnectSignal()
{
	client_t *cl;

	for (int i=0; i<MAXPLAYERS; i++)
	{
		cl = &clients[i];
		
		if (cl->cls < cs_spawned)
			continue;

		MSG_WriteByte(&cl->reliablebuf, svc_reconnect);
		SV_SendPacket(cl);
	   
		playeringame[i] = false;
		cl->cls = cs_free;
	   
		players[i].mo->Destroy();
		players[i].mo = NULL;
	}

}


void SV_SendDisconnectSignal()
{
	client_t *cl;

	for (int i=0; i<MAXPLAYERS; i++)
	{
		cl = &clients[i];
		
		if (cl->cls < cs_spawned)
			continue;
		
		MSG_WriteByte(&cl->reliablebuf, svc_disconnect);
		SV_SendPacket(cl);
	   
		playeringame[i] = false;
		cl->cls = cs_free;
	   
		players[i].mo->Destroy();
		players[i].mo = NULL;
	}

}

//
// SV_ExitLevel
// Called when timelimit or fraglimit hit.
//
void SV_ExitLevel()
{
	int		leadplayer;
	

	for (int i=0; i<MAXPLAYERS; i++)
	{
	   client_t *cl = &clients[i];

	   players[i].spectator = 10;

	   if (cl->cls < cs_spawned)
		   continue;

	   leadplayer = SV_GetLeadPlayer(&players[i]);

	   SV_CheckBuffer(cl, &cl->reliablebuf, 2);
	   MSG_WriteByte(&cl->reliablebuf, svc_exitlevel);
	   MSG_WriteByte(&cl->reliablebuf, leadplayer);
	}
}

//
// SV_BroadcastPrintf
// Sends text to all active clients.
//

#define MSGBUFSIZE 1024

void STACK_ARGS SV_BroadcastPrintf (int level, const char *fmt, ...)
{
    va_list       argptr;
    char          string[MSGBUFSIZE];
    client_t     *cl;

    va_start (argptr,fmt);
#ifdef _MSC_VER
	_vsnprintf (string, MSGBUFSIZE-1, fmt, argptr);
#else
	vsnprintf (string, MSGBUFSIZE-1, fmt, argptr);
#endif
    va_end (argptr);

    if (players[parse_cl].spectator > 0)
		Printf (level, "(SPEC) %s", string);  // print to the console
	else
		Printf (level, "%s", string);
	


    for (int i=0; i<MAXPLAYERS; i++)
    {
		cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;        

		// This chat string came from a spectator, only other spectators
		// can read this
		if (level == PRINT_CHAT && players[parse_cl].spectator > 0 &&
			players[i].spectator <= 0)
			continue;


		SV_CheckBuffer(cl, &cl->reliablebuf, 2+sizeof(string));
		MSG_WriteByte (&cl->reliablebuf, svc_print);
		MSG_WriteByte (&cl->reliablebuf, level);
		MSG_WriteString (&cl->reliablebuf, string);
    }
}


//
// SV_Say
// Show a chat string and send it to others clients.
//
void SV_Say(void)
{
	char *s = MSG_ReadString();
	
	// flood protection
	if (gametic - clients[parse_cl].lastsaytic > 15)
	{
		SV_BroadcastPrintf (PRINT_CHAT, "%s: %s\n", players[parse_cl].userinfo.netname, s);
		clients[parse_cl].lastsaytic = gametic;
	}

	clients[parse_cl].bytesin += sizeof(s);
}

//
// SV_UpdateMissiles
// Updates missiles position sometimes.
//
void SV_UpdateMissiles(client_t* cl)
{
    AActor *mo;

    TThinkerIterator<AActor> iterator;
    while ( (mo = iterator.Next() ) )
    {
        if (!(mo->flags & MF_MISSILE) )
			continue;

		if (mo->type == MT_PLASMA)
			continue;
            
		// update missile position every 50 tics
		if ( ((gametic + mo->netid) % 50) )
			continue;

		SV_CheckBuffer(cl, &cl->netbuf, 17);
		MSG_WriteByte (&cl->netbuf, svc_movemobj);
		MSG_WriteShort (&cl->netbuf, mo->netid);
		MSG_WriteLong (&cl->netbuf, mo->x);
		MSG_WriteLong (&cl->netbuf, mo->y);
		MSG_WriteLong (&cl->netbuf, mo->z);
    }


}


#define MAXCORPSES 15

//
// SV_RemoveCorpses
// Removes some corpses
//
void SV_RemoveCorpses (void)
{
	/*AActor *mo;
	int     corpses = 0;

	if (gametic%35)
	{
		return;
	}
	else
	{
		TThinkerIterator<AActor> iterator;
		while ( (mo = iterator.Next() ) )
		{
			if (mo->type == MT_PLAYER && (!mo->player || mo->health <=0) )
				corpses++;
		}
	}

	if (corpses > MAXCORPSES)
	{
		TThinkerIterator<AActor> iterator;
		while (corpses > MAXCORPSES && (mo = iterator.Next() ) )
		{
			if (mo->type == MT_PLAYER && !mo->player)
			{
				mo->Destroy();
				corpses--;
			}
		}
	}*/
}


//
// SV_SendGametic
// Sends gametic to calculate ping
//
void SV_SendGametic(client_t* cl)
{
	AActor	*mo;
	int		thingcount=0;
	
	if ((gametic%100) != 0)
		return;

	SV_CheckBuffer(cl, &cl->reliablebuf, 9);
	MSG_WriteByte (&cl->reliablebuf, svc_svgametic);
	MSG_WriteLong (&cl->reliablebuf, gametic);

	// [ZDaemon] - Report the number of items and report that too
	if (gamestate != GS_LEVEL)
	{	MSG_WriteLong (&cl->reliablebuf, 0);
		return;
	}

	TThinkerIterator<AActor> iterator;
	while ( (mo = iterator.Next() ) )
	{
		if ((mo->flags & (MF_MISSILE | MF_SHOOTABLE | MF_COUNTKILL | MF_SPECIAL)))
			thingcount++;
	}

	MSG_WriteLong (&cl->reliablebuf, thingcount);

}

// calculates ping using gametic which was sent by SV_SendGametic and
// current gametic
void SV_CalcPing(void)
{	
	int oldtic = MSG_ReadLong();

	int ping = gametic - oldtic;
	if (ping > 255) ping = 255;
	players[parse_cl].ping = ping;
}

//
// SV_UpdatePing
// send pings to a client
//
void SV_UpdatePing(client_t* cl)
{
	if ((gametic%101) != 0)
		return;

	for (int j=0; j<MAXPLAYERS; j++)
	{
		if (cl->cls < cs_spawned)
			continue;

	    SV_CheckBuffer(cl, &cl->reliablebuf, 3);
		MSG_WriteByte(&cl->reliablebuf, svc_updateping);
	    MSG_WriteByte(&cl->reliablebuf, j);  // player
	    MSG_WriteByte(&cl->reliablebuf, players[j].ping);
	}
}


//
// SV_UpdateDeadPlayers
// Update player's frame while he's dying.
//
void SV_UpdateDeadPlayers()
{
    AActor *mo;

    TThinkerIterator<AActor> iterator;
    while ( (mo = iterator.Next() ) )
    {
        if (mo->type != MT_PLAYER || mo->player)
			continue;
		
		if (mo->oldframe != mo->frame)
			for (int i=0; i<MAXPLAYERS; i++)
			{
				client_t *cl = &clients[i];

				if (cl->cls < cs_spawned)
					continue;

				SV_CheckBuffer(cl, &cl->reliablebuf, 4);
				MSG_WriteByte (&cl->reliablebuf, svc_mobjframe);
				MSG_WriteShort (&cl->reliablebuf, mo->netid);
				MSG_WriteByte (&cl->reliablebuf, mo->frame);
			}
		
		mo->oldframe = mo->frame;
    }

}


//
// SV_UpdateSectors
// Update doors, moving floors, ceilings etc...
//
void SV_UpdateSectors(client_t* cl)
{
	for (int s=0; s<numsectors; s++)
	{ 
		sector_t* sec = &sectors[s];

		if (sec->floordata && sec->floordata->IsKindOf (RUNTIME_CLASS(DMover)))
		{
			SV_CheckBuffer(cl, &cl->netbuf, 5);
			MSG_WriteByte (&cl->netbuf, svc_floorheight);
			MSG_WriteShort (&cl->netbuf, s);
			MSG_WriteShort (&cl->netbuf, sec->floorheight>>FRACBITS);
		}
		
		if (sec->ceilingdata && sec->ceilingdata->IsKindOf (RUNTIME_CLASS(DMover)))
		{
			SV_CheckBuffer(cl, &cl->netbuf, 5);
			MSG_WriteByte (&cl->netbuf, svc_ceilingheight);
			MSG_WriteShort (&cl->netbuf, s);
			MSG_WriteShort (&cl->netbuf, sec->ceilingheight>>FRACBITS);
		}

		//SV_CheckBuffer(cl, &cl->netbuf, 5);
		//MSG_WriteByte (&cl->netbuf, svc_lightlevel);
		//MSG_WriteShort (&cl->netbuf, s);
		//MSG_WriteShort (&cl->netbuf, sec->lightlevel);
		
	}
}

//
// SV_ClearClientsBPS
//
void SV_ClearClientsBPS(void)
{
	if (gametic % 35)
		return;

	for (int i=0; i<MAXPLAYERS; i++)
	{
		client_t *cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;

		if (clients[i].bytesin >= floodrate.value)
			SV_KickPlayer(i, "you've been disconnected for flooding.");

		cl->reliable_bps = 0;
		cl->unreliable_bps = 0;
		cl->bytesin = 0;
	}
}


//
// SV_WriteCommands
//
void SV_WriteCommands(void)
{
	int        i, j;
	client_t  *cl;

	
	SV_UpdateDeadPlayers(); // Update dying players.

	for (i=0; i<MAXPLAYERS; i++)
	{
		cl = &clients[i];

		if (cl->cls < cs_spawned)
			continue;
		
		// Don't need to update origin every tic.
		// The server sends origin and velocity of a 
		// player and the client always knows origin
		// on the next tic.
		if (gametic % 2)
			for (j=0; j<MAXPLAYERS; j++)
			{
				if (!playeringame[j] || j == i)
					continue;

				AActor *mobj = players[j].mo;

				if (!mobj)
					continue;

				byte bits = 0;

				// [NightFang] - I'm not exactly sure how many bytes
				// to resever here, I think 17 but 20 is a good 
				// round number to use
				SV_CheckBuffer(cl, &cl->netbuf, 20);
				MSG_WriteByte(&cl->netbuf, svc_moveplayer);
				MSG_WriteByte(&cl->netbuf, j);     // player number
				
				if (mobj->x != mobj->oldx)
					bits |= CM_X;
				if (mobj->y != mobj->oldy)
					bits |= CM_Y;
				if (mobj->z != mobj->oldz)
					bits |= CM_Z;
				if (mobj->angle != mobj->oldangle)
					bits |= CM_ANGLE;
				if (mobj->frame != mobj->oldframe)
					bits |= CM_FRAME;
				if (mobj->momx != mobj->oldmomx)
					bits |= CM_MOMX;
				if (mobj->momy != mobj->oldmomy)
					bits |= CM_MOMY;
				if (mobj->momz != mobj->oldmomz)
					bits |= CM_MOMZ;

				MSG_WriteByte(&cl->netbuf, bits);

				if (mobj->x != mobj->oldx)
					MSG_WriteLong(&cl->netbuf, mobj->x);				
				if (mobj->y != mobj->oldy)
					MSG_WriteLong(&cl->netbuf, mobj->y);
				if (mobj->z != mobj->oldz)
					MSG_WriteLong(&cl->netbuf, mobj->z);
				
				if (mobj->angle != mobj->oldangle)
					MSG_WriteLong(&cl->netbuf, mobj->angle);

				if (mobj->frame != mobj->oldframe)
				{
					if (mobj->frame == 32773)
						MSG_WriteByte(&cl->netbuf, PLAYER_FULLBRIGHTFRAME);
					else
						MSG_WriteByte(&cl->netbuf, mobj->frame);
				}

				// write velocity
				if (mobj->momx != mobj->oldmomx)
					MSG_WriteLong(&cl->netbuf, mobj->momx);
				if (mobj->momy != mobj->oldmomy)
					MSG_WriteLong(&cl->netbuf, mobj->momy);
				if (mobj->momz != mobj->oldmomz)
					MSG_WriteLong(&cl->netbuf, mobj->momz);
			}


		SV_UpdateConsolePlayer(cl);

		SV_UpdateMissiles(cl);

        SV_UpdateSectors(cl); // update doors, moving floors, ceilings etc		                        

	 	SV_SendGametic(cl); // used to calculate a ping value,
		                    // when a client returns it
     	SV_UpdatePing(cl);
	}

	// save players' origin, velocity, frame and angle
	if (gametic % 2)
		for (i=0; i<MAXPLAYERS; i++)
		{
			if (!playeringame[i])
				continue;

			AActor *mobj = players[i].mo;

			if (!mobj)
				continue;
		
			mobj->oldx = mobj->x;
			mobj->oldy = mobj->y;
			mobj->oldz = mobj->z;
			mobj->oldangle = mobj->angle;
			mobj->oldframe = mobj->frame;
			mobj->oldmomx = mobj->momx;
			mobj->oldmomy = mobj->momy;
			mobj->oldmomz = mobj->momz;
		}
}

void SV_PlayerTriedToCheat()
{
	SV_BroadcastPrintf(PRINT_HIGH, "%s tried to cheat!\n", players[parse_cl].userinfo.netname);
	SV_DropClient();
}

/*
struct usercmd_s {
	byte	msec;			// not sure how to use this yet...
	byte	buttons;
	short	pitch;			// up/down. currently just a y-sheering amount
	short	yaw;			// left/right	// If you haven't guessed, I just
	short	roll;			// tilt			// ripped these from Quake2's usercmd.
	short	forwardmove;
	short	sidemove;
	short	upmove;
	byte	impulse;
	byte	use;
};
*/

void SV_GetPlayerCmd(void)
{
	client_t *cl = &clients[parse_cl];
	ticcmd_t *cmd = &players[parse_cl].cmd;
	int       tic;

    tic = MSG_ReadLong(); // get client gametic

	// get the previous cmds and angle
	if (  cl->lastcmdtic && (tic - cl->lastcmdtic >= 2) 
		&& gamestate != GS_INTERMISSION )
	{
		cmd->ucmd.buttons = MSG_ReadByte();
	    players[parse_cl].mo->angle = MSG_ReadShort() << 16;
		
		if (dmflags & DF_NO_FREELOOK)
		{
			players[parse_cl].mo->pitch = 0;
			MSG_ReadShort();
		}
		else
			players[parse_cl].mo->pitch = MSG_ReadShort() << 16;

	    cmd->ucmd.forwardmove = MSG_ReadShort();
	    cmd->ucmd.sidemove = MSG_ReadShort();

		if ( abs(cmd->ucmd.forwardmove) > 12800 
			|| abs(cmd->ucmd.sidemove) > 12800)
		{
			SV_PlayerTriedToCheat();
			return;
		}

		cmd->ucmd.impulse = MSG_ReadByte();

		P_PlayerThink (&players[parse_cl]);
	    players[parse_cl].mo->RunThink();
	}
	else
	{
		// get 10 bytes
		MSG_ReadLong();
		MSG_ReadLong();
		MSG_ReadShort();
	}

	// get current cmds and angle
	cmd->ucmd.buttons = MSG_ReadByte();
	if (gamestate != GS_INTERMISSION)
	{
		players[parse_cl].mo->angle = MSG_ReadShort() << 16;
		
		if (dmflags & DF_NO_FREELOOK)
		{
			players[parse_cl].mo->pitch = 0;
			MSG_ReadShort();
		}
		else
			players[parse_cl].mo->pitch = MSG_ReadShort() << 16;
	}
	else
		MSG_ReadLong();

	cmd->ucmd.forwardmove = MSG_ReadShort();
	cmd->ucmd.sidemove = MSG_ReadShort();
	
	if ( abs(cmd->ucmd.forwardmove) > 12800 
		|| abs(cmd->ucmd.sidemove) > 12800)
	{
		SV_PlayerTriedToCheat();
		return;
	}
	
	cmd->ucmd.impulse = MSG_ReadByte();

	if (gamestate != GS_INTERMISSION)
	{
		P_PlayerThink (&players[parse_cl]);
		players[parse_cl].mo->RunThink();
	}

	cl->lastcmdtic = tic;
}


void SV_UpdateConsolePlayer(client_t* cl)
{
	int        playernum;
	player_t  *player;

	// It's not a good idea to send 33 bytes every tic.
	if (gametic % 3)
		return;

	for (playernum=0; playernum<MAXPLAYERS; playernum++)
	{
		if (cl->cls < cs_spawned)
			continue;

		if (cl == &clients[playernum])
		{
			player = &players[playernum];
			break;
		}
	}
	
	if (!player)
	    return;
	
	// haven't got any cmds yet
	if (!cl->lastcmdtic)
		return;

	// client player will update his position if packets were missed
	SV_CheckBuffer(cl, &cl->netbuf, 36);
	MSG_WriteByte (&cl->netbuf, svc_updatelocalplayer);
	MSG_WriteLong (&cl->netbuf, cl->lastcmdtic);

	MSG_WriteLong (&cl->netbuf, player->mo->x);
	MSG_WriteLong (&cl->netbuf, player->mo->y);
	MSG_WriteLong (&cl->netbuf, player->mo->z);

	MSG_WriteLong (&cl->netbuf, player->mo->momx);
	MSG_WriteLong (&cl->netbuf, player->mo->momy);
	MSG_WriteLong (&cl->netbuf, player->mo->momz);

//	MSG_WriteShort (&cl->netbuf, player->mo->momx >> FRACBITS);
//	MSG_WriteShort (&cl->netbuf, player->mo->momy >> FRACBITS);
//	MSG_WriteShort (&cl->netbuf, player->mo->momz >> FRACBITS);
}

void SV_UpdateThing(client_t	*cl)
{
	AActor	*mo;
	int		handle = MSG_ReadShort();	// monster id
	
	TThinkerIterator<AActor> iterator;
    while ( (mo = iterator.Next() ) )
    {
		if (handle == mo->handle_id)
		{
			if ((mo->flags & (MF_SHOOTABLE | MF_SPECIAL | MF_COUNTKILL | MF_MISSILE)))
			{
				// The thing exists and its update-able. Heh
				SV_CheckBuffer(cl, &cl->netbuf, 3);
				MSG_WriteByte(&cl->reliablebuf, svc_updatething);
				MSG_WriteShort(&cl->reliablebuf, handle);
				MSG_WriteByte(&cl->reliablebuf, 10);
				return;
			}
		}
	}

	// Thing not found
	SV_CheckBuffer(cl, &cl->netbuf, 3);
	MSG_WriteByte(&cl->reliablebuf, svc_updatething);
	MSG_WriteShort(&cl->reliablebuf, handle);
	MSG_WriteByte(&cl->reliablebuf, 0);
}


void SV_SendNetIDs(client_t*	cl)
{
	AActor	*mo;

	TThinkerIterator<AActor> iterator;
    while ( (mo = iterator.Next() ) )
	{
		if ((mo->flags & ( MF_MISSILE | MF_SHOOTABLE | MF_COUNTKILL | MF_SPECIAL)))
		{
			SV_CheckBuffer(cl, &cl->netbuf, 3);
			MSG_WriteByte(&cl->reliablebuf, svc_updatethingcount);
			MSG_WriteShort(&cl->reliablebuf, mo->handle_id);
		}
	}
}




AActor	*SV_FindThingByHandle(int	num)
{

	AActor	*mo;

	TThinkerIterator<AActor> iterator;
    while ( (mo = iterator.Next() ) )
	{
		if (mo->handle_id == num)
			return mo;
	}

	return 0;

}


//
// SV_FixClientThing - The client is missing this thing, so spawn it
//

void	SV_FixClientThing(client_t*	cl, int	num)    
{
	AActor	*mo = SV_FindThingByHandle(num);

	if (!mo)
		return;

	SV_CheckBuffer(cl, &cl->reliablebuf, 20);
	MSG_WriteByte(&cl->reliablebuf, svc_spawnmobj);
	MSG_WriteLong(&cl->reliablebuf, mo->x);
	MSG_WriteLong(&cl->reliablebuf, mo->y);
	MSG_WriteLong(&cl->reliablebuf, mo->z);
	MSG_WriteShort(&cl->reliablebuf, mo->type);
	MSG_WriteShort(&cl->reliablebuf, mo->netid);
	// [NightFang] - Added this for ghosts
	MSG_WriteShort(&cl->reliablebuf, mo->handle_id);

	/*SV_CheckBuffer(cl, &cl->reliablebuf, 5);
	MSG_WriteByte (&cl->reliablebuf, svc_mobjstate);
	MSG_WriteShort (&cl->reliablebuf, mo->netid);
	MSG_WriteShort (&cl->reliablebuf, mo->state);*/
}

// Function prototype
void G_DoReborn (int playernum);

void SV_SpectatorUpdate(int	num)
{
	client_t*	cl = &clients[num];
	BOOL		domessage = true;
	int			numplayers=0;
	int			team;
	int			client_numlines;	// lines checking
	int			client_numsectors;	// sectors checking

	if (players[num].spectator <= 0)
		domessage = false;

	cl->lastcmdtic = MSG_ReadLong();

	/*players[num].cmd.ucmd.buttons = MSG_ReadByte();
	if (gamestate != GS_INTERMISSION)
	{
		players[num].mo->angle = MSG_ReadShort() << 16;
		
		if (dmflags & DF_NO_FREELOOK)
		{
			players[num].mo->pitch = 0;
			MSG_ReadShort();
		}
		else
			players[num].mo->pitch = MSG_ReadShort() << 16;
	}
	else
		MSG_ReadLong();

	players[num].cmd.ucmd.forwardmove = MSG_ReadShort();
	players[num].cmd.ucmd.sidemove = MSG_ReadShort();
	players[num].cmd.ucmd.impulse = MSG_ReadByte();*/

	client_numlines = MSG_ReadByte();
	client_numsectors = MSG_ReadByte();

	if (client_numlines - numlines > mapcheckdifference.value ||
		client_numsectors - numsectors > mapcheckdifference.value)
	{
		SV_KickPlayer(num, "i dont trust that map your client is using");
		return;
	}

	//if (gamestate != GS_INTERMISSION)
	//{
	//	P_PlayerThink (&players[num]);
	//	players[num].mo->RunThink();
	//}


	// Time to respawn!
	//if (players[num].cmd.ucmd.buttons & BT_ATTACK)
	//{

	// Make sure its okay first
	for (int i=0; i<maxclients; i++)
	{
		if (clients[i].cls < cs_spawned)
			continue;

		if (players[i].spectator > 0) // spectators dont count
			continue; 
		numplayers++; // get current number of players
	}

		// Have map error checking here!
		// This is the best spot to put it.  Player cant even join
		// if their using a maphack, right?

	if (numplayers >= max_players)
	{  
		MSG_WriteByte(&clients[num].reliablebuf, svc_cantjoin);
		return;
	}

	if (!deathmatch.value && !teamplay.value)
	{
		players[num].spectator = 0;
		players[num].cheats &= ~CF_NOTARGET;
		G_DoReborn(num);
	}
	else
	{ 
		players[num].playerstate = PST_REBORN; 
		players[num].spectator = 0;
		players[num].cheats &= ~CF_NOTARGET;
	}

		
	if (teamplay.value)
	{

		team = SV_ChooseTeamForPlayer();
	
		// Be sure
		if (team == 0)
			strcpy(players[parse_cl].userinfo.team, "RED");
		else
			strcpy(players[parse_cl].userinfo.team, "BLUE");
	
		// Send special settings like teamplay and deathmatch.
		MSG_WriteByte(&cl->reliablebuf, svc_specialsettings);
		MSG_WriteByte(&cl->reliablebuf, (byte)teamplay.value);
		MSG_WriteByte(&cl->reliablebuf, team);
		MSG_WriteByte(&cl->reliablebuf, (byte)deathmatch.value);
	}

	if (!domessage) 
		return;


	if (!teamplay.value)
	{
		SV_BroadcastPrintf (PRINT_HIGH, "%s has joined the game.\n", players[parse_cl].userinfo.netname);
	}
	else
	{ 
		if (team == 0)
			SV_BroadcastPrintf (PRINT_HIGH, "%s has joined the game and joined the RED team.\n", players[parse_cl].userinfo.netname);
		if (team == 1)
			SV_BroadcastPrintf (PRINT_HIGH, "%s has joined the game and joined the BLUE team.\n", players[parse_cl].userinfo.netname);
	}

}


void SV_ClearSpectatorUpdate(void)
{
	MSG_ReadLong();

	MSG_ReadByte();
	if (gamestate != GS_INTERMISSION)
	{
		MSG_ReadShort();
		MSG_ReadShort();
		
	}
	else
		MSG_ReadLong();

	MSG_ReadShort();
	MSG_ReadShort();
	MSG_ReadByte();
}

	




//
// SV_ParseCommands
//
void SV_ParseCommands(void)
{		
	 int  cmd;
	 int  specupdate=0; // Will only process the clc_spectatorupdate
						// once perpacket
	 char *str;
	 
	 while(1)	 
	 {  
 		if (msg_badread)
		{
            Printf (PRINT_HIGH, "SV_ParseCommands: badread\n");
          	SV_DropClient();
            break;
		}

		cmd = MSG_ReadByte();

		if (cmd == -1) // EOM
			break;

		if (clients[parse_cl].cls < cs_spawned)
			break;

	    
		switch(cmd)
		{
		// For users who trail in their connecting abilities
		case S2C_CONNECTION:
			SV_SpawnClient();
			break;

		case clc_spectatorupdate:
			SV_SpectatorUpdate(parse_cl);
			break;

		case clc_getmissedthing:
			SV_FixClientThing(&clients[parse_cl], MSG_ReadShort());
			break;

		case clc_updatething:
			SV_UpdateThing(&clients[parse_cl]);
			break;

		case clc_updatethingcount:
			SV_SendNetIDs(&clients[parse_cl]);
			break;

		case clc_disconnect:
			SV_DisconnectClient();
			return;

        case clc_userinfo:
			SV_SetupUserInfo();
			break;

		case clc_say:
			SV_Say();
			break;

		case clc_move:
			SV_GetPlayerCmd();
			break;

		case clc_svgametic:
			SV_CalcPing();
			break;

		case clc_rate:
			clients[parse_cl].rate = MSG_ReadLong();	
			break;

		case clc_ack:
			SV_AcknowledgePacket();
			break;
	
		case clc_rcon:
			str = MSG_ReadString();
			if (clients[parse_cl].allow_rcon)
				AddCommandString (str);
			break;

		case clc_rcon_password:
			str = MSG_ReadString();
			if (strlen(rcon_password.string) >= 5 && !strcmp(rcon_password.string, str) )
				clients[parse_cl].allow_rcon = true;
			break;
		
		default:			
			Printf(PRINT_HIGH, "(%s) Unknown client message: %d\n", players[parse_cl].userinfo.netname, cmd);
			SZ_Clear(&clients[parse_cl].reliablebuf);
			SZ_Clear(&clients[parse_cl].netbuf);
            return;
		}	 
	}
}


//
// SV_RunTics
//
void SV_RunTics (void)
{
	int              nowtime;
	int              newtics;

	I_DoSelect();

	nowtime = I_GetTime ();
        newtics = nowtime - gametime;

	// console input
	char *cmd = I_ConsoleInput();
	if (cmd)
		AddCommandString (cmd);

	SV_GetPackets(); // get packets

	if (!(gametic % 35))
	{
		if (server_outdated > 0)
		{
			Printf(PRINT_HIGH, "\n\n***ATTENTION***\n");
			Printf(PRINT_HIGH, "YOUR SERVER IS OUT-DATED! PLEASE UPGRADE YOUR SERVER TO THE\n");
			Printf(PRINT_HIGH, "NEWEST VERSION OF ZDAEMON %d.%d.\n\n", server_outdated/100, server_outdated % 100);
		}
	}

	//SV_EnforceBans();

	if (newtics > 0)
	{
		DObject::BeginFrame ();

		// run the newtime tics
		while (newtics--)
		{
			C_Ticker ();
			G_Ticker ();
			SV_UpdateMaster();
			
			gametic++;
		}

		DObject::EndFrame ();
	
		gametime = nowtime;
	}
}



//////// FRAGS LOGGING ///////////

FILE  *sv_fraglogfile;

//
// SV_Fraglogfile
// from quake world
//
void SV_Fraglogfile (void)
{
	char	name[32];
	int		i;

	if (sv_fraglogfile)
	{
		Printf (PRINT_HIGH, "Frag file logging off.\n");
		fclose (sv_fraglogfile);
		sv_fraglogfile = NULL;
		return;
	}

	// find an unused name
	for (i=0 ; i<1000 ; i++)
	{
		sprintf (name, "frag_%i.log", i);
		sv_fraglogfile = fopen (name, "r");
		if (!sv_fraglogfile)
		{	// can't read it, so create this one
			sv_fraglogfile = fopen (name, "w");
			if (!sv_fraglogfile)
				i=1000;	// give error
			break;
		}
		fclose (sv_fraglogfile);
	}

	if (i==1000)
	{
		Printf (PRINT_HIGH, "Can't open any logfiles.\n");
		sv_fraglogfile = NULL;
		return;
	}

	Printf (PRINT_HIGH, "Logging frags to %s.\n", name);
}


//
// SV_logfrag
//
void SV_Logfrag (player_t* source, player_t* target)
{
	if (!sv_fraglogfile) 
		return;

	fprintf (sv_fraglogfile, "\\%s\\%s\\\n",
		source->userinfo.netname, target->userinfo.netname);
	fflush (sv_fraglogfile);
}






//////// CONNECTION LOGGING ///////////

FILE  *sv_connectlogfile;

//
// SV_Fraglogfile
// from quake world
//
void SV_Connectlogfile (void)
{
	char	name[32];
	int		i;

	if (sv_connectlogfile)
	{
		Printf (PRINT_HIGH, "Connect file logging off.\n");
		fclose (sv_connectlogfile);
		sv_connectlogfile = NULL;
		return;
	}

	// find an unused name
	for (i=0 ; i<1000 ; i++)
	{
		sprintf (name, "conn_%i.log", i);
		sv_connectlogfile = fopen (name, "r");
		if (!sv_connectlogfile)
		{	// can't read it, so create this one
			sv_connectlogfile = fopen (name, "w");
			if (!sv_connectlogfile)
				i=1000;	// give error
			break;
		}
		fclose (sv_connectlogfile);
	}

	if (i==1000)
	{
		Printf (PRINT_HIGH, "Can't open any connect log files.\n");
		sv_connectlogfile = NULL;
		return;
	}

	Printf (PRINT_HIGH, "Logging connections to %s.\n", name);
}


//
// SV_logconnection
//
void SV_Logconnection (char *text)
{
	if (!sv_connectlogfile) 
		return;

	fprintf (sv_connectlogfile, "%s\n",text);
	fflush (sv_connectlogfile);
}



BEGIN_COMMAND(connectlog)
{
	SV_Connectlogfile();
}
END_COMMAND(connectlog)