#include "gZDLaunch.h"

void ParseConfig(void)
{
	FILE *gcsdlrc;
	char configfile[1024];

	sprintf(configfile, "%s/.gzdl", getenv("HOME"));
	gcsdlrc = fopen(configfile, "r");

	strcpy(Config.Files.WadDir, WADDIR);
	strcpy(Config.Files.doomcl, ZDAEMON);
	strcpy(Config.Files.Iwad1, IWAD1);
	strcpy(Config.Files.Iwad2, IWAD2);
	strcpy(Config.Files.Iwadhc, IWADHC);
	strcpy(Config.Files.Iwadhx, IWADHX);
	strcpy(Config.Files.Iwadp, IWADP);
	strcpy(Config.Files.Iwadt, IWADT);

	if (!gcsdlrc)
	{
		PrintConsole
		    ("Could not open config file ~/.gzdl ... will use defaults\n",
		     "error");
		return;
	}

	if (!feof(gcsdlrc))
	{
		do
		{
			CFGLine cfg;

			cfg.value = malloc(sizeof(char) * (255 - 32));
			cfg.keyword = malloc(sizeof(char) * 32);

			fgets(cfg.wholeline, 255, gcsdlrc);

			if (strlen(cfg.wholeline) < 5)
				continue;
			delchar(cfg.wholeline, ' ');
			delchar(cfg.wholeline, '\t');
			delchar(cfg.wholeline, '\n');
			if (!strncmp(cfg.wholeline, "#", 1))
				continue;

			cfg.keyword = strtok(cfg.wholeline, "=");
			cfg.value = strtok(NULL, "=");

			if (!cfg.value)
				continue;

			configure(cfg.keyword, cfg.value);

		}
		while (!feof(gcsdlrc));
	}
	fclose(gcsdlrc);
}

void configure(char *keyword, char *value)
{
	char tmp[255];

	if (!strcasecmp(keyword, "WadDir"))
		strcpy(Config.Files.WadDir, value);
	else if (!strcasecmp(keyword, "Iwad1"))
		strcpy(Config.Files.Iwad1, value);
	else if (!strcasecmp(keyword, "Iwad2"))
		strcpy(Config.Files.Iwad2, value);
	else if (!strcasecmp(keyword, "ZDaemon"))
		strcpy(Config.Files.doomcl, value);
	else if (!strcasecmp(keyword, "IwadHeretic"))
		strcpy(Config.Files.Iwadhc, value);
	else if (!strcasecmp(keyword, "IwadHexen"))
		strcpy(Config.Files.Iwadhx, value);
	else if (!strcasecmp(keyword, "IwadPlutonia"))
		strcpy(Config.Files.Iwadp, value);
	else if (!strcasecmp(keyword, "IwadTNT"))
		strcpy(Config.Files.Iwadt, value);
	else
	{
		sprintf(tmp, "%s: unknown keyword in config file\n",
			keyword);
		PrintConsole(tmp, "error");
	}
}

void save_config(void)
{
	FILE *gcsdlrc;
	char configfile[1024];

	sprintf(configfile, "%s/.gzdl", getenv("HOME"));
	gcsdlrc = fopen(configfile, "w");

	if (!gcsdlrc)
	{
		PrintConsole("Could not open config file for writing\n",
			     "error");
		return;
	}

	fprintf(gcsdlrc, "# This file was generated by gzdl %s\n",
		VERSION);
	fprintf(gcsdlrc, "# You may edit it by hand if you want :)\n\n");
	fprintf(gcsdlrc, "WadDir = %s\n", Config.Files.WadDir);
	fprintf(gcsdlrc, "Iwad1 = %s\n", Config.Files.Iwad1);
	fprintf(gcsdlrc, "Iwad2 = %s\n", Config.Files.Iwad2);
	fprintf(gcsdlrc, "IwadHexen = %s\n", Config.Files.Iwadhx);
	fprintf(gcsdlrc, "IwadHeretic = %s\n", Config.Files.Iwadhc);
	fprintf(gcsdlrc, "IwadPlutonia = %s\n", Config.Files.Iwadp);
	fprintf(gcsdlrc, "IwadTNT = %s\n", Config.Files.Iwadt);
	fprintf(gcsdlrc, "zdaemon = %s\n", Config.Files.doomcl);
}

void SaveNotify(void)
{
	int ret;
	int i = 0;
	char *sentry;
	FILE *notifyptr;
	char notifyfile[255];


	sprintf(notifyfile, "%s/.gzdl-notify", getenv("HOME"));

	notifyptr = fopen(notifyfile, "w");

	if (!notifyptr)
	{
		return;
	}

	do
	{
		ret =
		    gtk_clist_get_text(GTK_CLIST(notifylist), i,
				       0, &sentry);

		if (!ret)
		{
			fclose(notifyptr);
			return;
		}

		fprintf(notifyptr, "%s\n", sentry);
		i++;
	}
	while (ret);

	fclose(notifyptr);
}

void ApplyChanges(void)
{
	strcpy(Config.Files.WadDir,
	       gtk_entry_get_text(GTK_ENTRY(waddirentry)));
	strcpy(Config.Files.Iwad1,
	       gtk_entry_get_text(GTK_ENTRY(iwad1entry)));
	strcpy(Config.Files.Iwad2,
	       gtk_entry_get_text(GTK_ENTRY(iwad2entry)));
	strcpy(Config.Files.doomcl,
	       gtk_entry_get_text(GTK_ENTRY(doomclentry)));
	strcpy(Config.Files.Iwadhc,
	       gtk_entry_get_text(GTK_ENTRY(iwadhcentry)));
	strcpy(Config.Files.Iwadhx,
	       gtk_entry_get_text(GTK_ENTRY(iwadhxentry)));
	strcpy(Config.Files.Iwadp,
	       gtk_entry_get_text(GTK_ENTRY(iwadpentry)));
	strcpy(Config.Files.Iwadt,
	       gtk_entry_get_text(GTK_ENTRY(iwadtentry)));
	opened_config = 0;
	SaveNotify();
}

void ConfigDialog(void)
{
	GtkWidget *cdialog;
	GtkWidget *notebook;
	GtkWidget *tablenotify;
	GtkWidget *tablefiles;
	GtkWidget *scroll;
	GtkWidget *button;
#ifdef USE_GNOME
	GtkWidget *doomclbrowse;
	GtkWidget *iwad1browse;
	GtkWidget *iwad2browse;
	GtkWidget *iwadhxbrowse;
	GtkWidget *iwadhcbrowse;
	GtkWidget *iwadpbrowse;
	GtkWidget *iwadtbrowse;
	GtkWidget *waddirbrowse;
#endif
	GtkWidget *maintable;
	char *titles[1] = { "Name" };

	if (opened_config == 1)
		return;

	opened_config = 1;


	cdialog = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(cdialog), "Configuration");

	notebook = gtk_notebook_new();

	tablenotify = gtk_table_new(2, 2, FALSE);
	tablefiles = gtk_table_new(2, 8, FALSE);
	maintable = gtk_table_new(1, 2, FALSE);

	notifylist = gtk_clist_new_with_titles(1, titles);
	gtk_clist_set_auto_sort(GTK_CLIST(notifylist), TRUE);
	ReadNotifyToConfigDialog(GTK_WIDGET(notifylist));

	scroll = gtk_scrolled_window_new(NULL, NULL);

	gtk_container_add(GTK_CONTAINER(cdialog), maintable);
	gtk_container_add(GTK_CONTAINER(scroll), notifylist);

	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_AUTOMATIC);

	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tablefiles,
				 gtk_label_new("Files"));
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tablenotify,
				 gtk_label_new("Notify List"));


	gtk_table_attach(GTK_TABLE(tablenotify), scroll, 1, 2, 0, 2,
			 GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 2,
			 2);

	button = gtk_button_new_with_label("Delete");
	gtk_table_attach(GTK_TABLE(tablenotify), button, 0, 1, 0, 1,
			 GTK_FILL, GTK_SHRINK, 2, 2);

	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
				  GTK_SIGNAL_FUNC(DeleteNotifyEntry),
				  NULL);

	button = gtk_button_new_with_label("Add");
	gtk_table_attach(GTK_TABLE(tablenotify), button, 0, 1, 1, 2,
			 GTK_FILL, GTK_SHRINK, 2, 2);

	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
				  GTK_SIGNAL_FUNC(AddNotifyDialog), NULL);
	gtk_signal_connect(GTK_OBJECT(notifylist), "select_row",
			   GTK_SIGNAL_FUNC(SelectNotify), NULL);

#ifdef USE_GNOME
	doomclbrowse =
	    gnome_file_entry_new(NULL, "Pick Zdaemon executable");
	iwad1browse =
	    gnome_file_entry_new(NULL, "Pick Ultimate Doom IWAD");
	iwad2browse = gnome_file_entry_new(NULL, "Pick Doom ][ IWAD");
	iwadhcbrowse = gnome_file_entry_new(NULL, "Pick a Heretic IWAD");
	iwadhxbrowse = gnome_file_entry_new(NULL, "Pick a Hexen IWAD");
	iwadpbrowse =
	    gnome_file_entry_new(NULL, "Pick a Final Doom Plutonia IWAD");
	iwadtbrowse =
	    gnome_file_entry_new(NULL, "Pick a Final Doom TNT IWAD");

	waddirbrowse = gnome_file_entry_new(NULL, "Pick a wad directory");
	gnome_file_entry_set_directory(GNOME_FILE_ENTRY(waddirbrowse),
				       TRUE);

	doomclentry =
	    gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(doomclbrowse));
	iwad1entry =
	    gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(iwad1browse));
	iwad2entry =
	    gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(iwad2browse));
	iwadhcentry =
	    gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(iwadhcbrowse));
	iwadhxentry =
	    gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(iwadhxbrowse));
	iwadpentry =
	    gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(iwadpbrowse));
	iwadtentry =
	    gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(iwadtbrowse));

	waddirentry =
	    gnome_file_entry_gtk_entry(GNOME_FILE_ENTRY(waddirbrowse));
#else
	doomclentry = gtk_entry_new();
	iwad1entry = gtk_entry_new();
	iwad2entry = gtk_entry_new();
	iwadhcentry = gtk_entry_new();
	iwadhxentry = gtk_entry_new();
	iwadpentry = gtk_entry_new();
	iwadtentry = gtk_entry_new();
	waddirentry = gtk_entry_new();

#endif
	gtk_table_attach(GTK_TABLE(tablefiles),
			 gtk_label_new("zdaemon executable:"), 0, 1, 0, 1,
			 GTK_FILL, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles),
			 gtk_label_new("Ultimate Doom IWAD:"), 0, 1, 1, 2,
			 GTK_FILL, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles),
			 gtk_label_new("Doom ][ IWAD:"), 0, 1, 2, 3,
			 GTK_FILL, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles),
			 gtk_label_new("Heretic IWAD:"), 0, 1, 3, 4,
			 GTK_FILL, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles),
			 gtk_label_new("Hexen IWAD:"), 0, 1, 4, 5,
			 GTK_FILL, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles),
			 gtk_label_new("Plutonia IWAD:"), 0, 1, 5, 6,
			 GTK_FILL, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles),
			 gtk_label_new("TNT IWAD:"), 0, 1, 6, 7,
			 GTK_FILL, GTK_SHRINK, 2, 2);

	gtk_table_attach(GTK_TABLE(tablefiles),
			 gtk_label_new("Wad directory:"), 0, 1, 7, 8,
			 GTK_FILL, GTK_SHRINK, 2, 2);

#ifdef USE_GNOME
	gtk_table_attach(GTK_TABLE(tablefiles), doomclbrowse, 1, 2, 0, 1,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwad1browse, 1, 2, 1, 2,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwad2browse, 1, 2, 2, 3,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), waddirbrowse, 1, 2, 7, 8,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwadhcbrowse, 1, 2, 3, 4,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwadhxbrowse, 1, 2, 4, 5,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwadpbrowse, 1, 2, 5, 6,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwadtbrowse, 1, 2, 6, 7,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);

	gtk_table_attach(GTK_TABLE(maintable), notebook, 0, 1, 0, 1,
			 GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 2,
			 2);
#else
	gtk_table_attach(GTK_TABLE(tablefiles), doomclentry, 1, 2, 0, 1,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwad1entry, 1, 2, 1, 2,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwad2entry, 1, 2, 2, 3,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), waddirentry, 1, 2, 7, 8,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwadhcentry, 1, 2, 3, 4,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwadhxentry, 1, 2, 4, 5,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwadpentry, 1, 2, 5, 6,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);
	gtk_table_attach(GTK_TABLE(tablefiles), iwadtentry, 1, 2, 6, 7,
			 GTK_FILL | GTK_EXPAND, GTK_SHRINK, 2, 2);

	gtk_table_attach(GTK_TABLE(maintable), notebook, 0, 1, 0, 1,
			 GTK_FILL | GTK_EXPAND, GTK_FILL | GTK_EXPAND, 2,
			 2);

#endif
	button = gtk_button_new_with_label("OK");
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
				  ApplyChanges, NULL);
//      gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
//                                SaveNotify, NULL);
	gtk_signal_connect_object(GTK_OBJECT(button), "clicked",
				  gtk_widget_destroy, GTK_OBJECT(cdialog));
	gtk_signal_connect_object(GTK_OBJECT(cdialog), "delete_event",
				  CloseConfig, NULL);
	gtk_signal_connect_object(GTK_OBJECT(cdialog), "delete_event",
				  gtk_widget_destroy, GTK_OBJECT(cdialog));
	gtk_table_attach(GTK_TABLE(maintable), button, 0, 1, 1, 2,
			 GTK_FILL, GTK_FILL, 2, 2);

	gtk_entry_set_text(GTK_ENTRY(doomclentry), Config.Files.doomcl);
	gtk_entry_set_text(GTK_ENTRY(iwad1entry), Config.Files.Iwad1);
	gtk_entry_set_text(GTK_ENTRY(iwad2entry), Config.Files.Iwad2);
	gtk_entry_set_text(GTK_ENTRY(iwadhcentry), Config.Files.Iwadhc);
	gtk_entry_set_text(GTK_ENTRY(iwadhxentry), Config.Files.Iwadhx);
	gtk_entry_set_text(GTK_ENTRY(waddirentry), Config.Files.WadDir);
	gtk_entry_set_text(GTK_ENTRY(iwadpentry), Config.Files.Iwadp);
	gtk_entry_set_text(GTK_ENTRY(iwadtentry), Config.Files.Iwadt);


	gtk_widget_show_all(cdialog);
}

void CloseConfig(void)
{
	opened_config = 0;
}

void SelectNotify(GtkWidget * list, int row, int col)
{
	notify_row = row;
}

void DeleteNotifyEntry(void)
{
	gtk_clist_remove(GTK_CLIST(notifylist), notify_row);
}

void ReadNotifyToConfigDialog(GtkWidget * notifylist)
{
	char filename[255];
	char tmp[16];
	char *row[1];
	FILE *notifydb;

	sprintf(filename, "%s/.gzdl-notify", getenv("HOME"));

	notifydb = fopen(filename, "r");

	if (!notifydb)
	{
		PrintConsole
		    ("No notify list available, creating new one\n",
		     "error");

		notifydb = fopen(filename, "w");
		if (!notifydb)
			return;

		fprintf(notifydb, "slayer[nfc]\n");
		fprintf(notifydb, "RvB\n\n");
		fclose(notifydb);

		notifydb = fopen(filename, "r");

	}

	fgets(tmp, 16, notifydb);
	if (feof(notifydb))
		return;
	do
	{
		delchar(tmp, '\n');
		delchar(tmp, '\t');
		if (strlen(tmp) < 1)
			return;
		row[0] = tmp;
		gtk_clist_append(GTK_CLIST(notifylist), row);
		fgets(tmp, 16, notifydb);
	}
	while (!feof(notifydb));

	fclose(notifydb);
}

void AddNotifyDialog(void)
{
	GtkWidget *dialog;
	GtkWidget *widget;
	GtkWidget *table;
	GtkWidget *addnotifier;
	dialog = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title(GTK_WINDOW(dialog),
			     "Add user to notify list");
	table = gtk_table_new(2, 2, TRUE);
	gtk_container_add(GTK_CONTAINER(dialog), table);
	gtk_container_set_border_width(GTK_CONTAINER(dialog), 10);
	PrintConsole
	    ("Please enter the name of a doom player to the dialog box and press OK\n",
	     "normal");

	widget = gtk_button_new_with_label("Cancel");
	gtk_table_attach_defaults(GTK_TABLE(table), widget, 1, 2, 1, 2);
	gtk_signal_connect_object(GTK_OBJECT(widget), "clicked",
				  GTK_SIGNAL_FUNC(gtk_widget_destroy),
				  GTK_OBJECT(dialog));
	widget = gtk_button_new_with_label("Add");
	addnotifier = gtk_entry_new();

	gtk_table_attach_defaults(GTK_TABLE(table), widget, 0, 1, 1, 2);
	gtk_signal_connect_object(GTK_OBJECT(widget), "clicked",
				  GTK_SIGNAL_FUNC(AddNotify),
				  GTK_OBJECT(addnotifier));
	gtk_signal_connect_object(GTK_OBJECT(widget), "clicked",
				  GTK_SIGNAL_FUNC(gtk_widget_destroy),
				  GTK_OBJECT(dialog));

	gtk_table_attach_defaults(GTK_TABLE(table), addnotifier, 0,
				  2, 0, 1);

	gtk_widget_show_all(dialog);
}

void AddNotify(GtkWidget * entry)
{
	char *row[1];
	row[0] = gtk_entry_get_text(GTK_ENTRY(entry));
	gtk_clist_append(GTK_CLIST(notifylist), row);
}
